// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.net.*;
import java.io.*;
import java.util.*;
import org.xwt.util.*;
import org.mozilla.javascript.*;
import org.bouncycastle.util.encoders.Base64;
import org.bouncycastle.crypto.digests.*;

/**
 *  This object encapsulates a *single* HTTP connection. Multiple requests may be pipelined over a connection (thread-safe),
 *  although any IOException encountered in a request will invalidate all later requests.
 */
public class HTTP {

    /** the URL as passed to the original constructor; this is never changed */
    final String originalUrl;

    /** the URL to connect to; this is munged when the url is parsed */
    URL url = null;

    /** the host to connect to */
    String host = null;

    /** the port to connect on */
    int port = -1;

    /** true if SSL (HTTPS) should be used */
    boolean ssl = false;

    /** the path (URI) to retrieve on the server */
    String path = null;

    /** the socket */
    Socket sock = null;

    /** the socket's inputstream */
    InputStream in = null;

    /** the username and password portions of the URL */
    String userInfo = null;

    /** cache of userInfo strings, keyed on originalUrl */
    private static Hashtable authCache = new Hashtable();

    /** this is null if the current request is the first request on
     *  this HTTP connection; otherwise it is a Semaphore which will be
     *  released once the request ahead of us has recieved its response
     */
    Semaphore okToRecieve = null;

    /** cache for resolveAndCheckIfFirewalled() */
    static Hashtable resolvedHosts = new Hashtable();

    /** if any request encounters an IOException, the entire HTTP connection is invalidated */
    boolean invalid = false;

    /** true iff we are allowed to skip the resolve check (only allowed when we're downloading the PAC script) */
    boolean skipResolveCheck = false;


    // Public Methods ////////////////////////////////////////////////////////////////////////////////////////

    public HTTP(String url) { this(url, false); }
    public HTTP(String url, boolean skipResolveCheck) {
        originalUrl = url;
        this.skipResolveCheck = skipResolveCheck;
    }

    /** Performs an HTTP GET request */
    public HTTPInputStream GET() throws IOException { return makeRequest(null, null); }

    /** Performs an HTTP POST request; content is appended to the headers (so it should include a blank line to delimit the beginning of the body) */
    public HTTPInputStream POST(String contentType, String content) throws IOException { return makeRequest(contentType, content); }

    /**
     *  This method isn't synchronized; however, only one thread can be in the inner synchronized block at a time, and the rest of
     *  the method is protected by in-order one-at-a-time semaphore lock-steps
     */
    private HTTPInputStream makeRequest(String contentType, String content) throws IOException {

        // Step 1: send the request and establish a semaphore to stop any requests that pipeline after us
        Semaphore blockOn = null;
        Semaphore releaseMe = null;
        synchronized(this) {
            if (invalid) throw new HTTPException("connection failed on a previous pipelined call");
            try {
                connect();
                sendRequest(contentType, content);
            } catch (IOException e) {
                invalid = true;
                throw e;
            }
            blockOn = okToRecieve;
            releaseMe = okToRecieve = new Semaphore();
        }
        
        // Step 2: wait for requests ahead of us to complete, then read the reply off the stream
        boolean doRelease = true;
        try {
            if (blockOn != null) blockOn.block();
            if (invalid) throw new HTTPException("connection failed on a previous pipelined call");
            
            Hashtable h = in == null ? null : parseHeaders(in);
            if (h == null) {
                // sometimes the server chooses to close the stream between requests
                in = null; sock = null;
                releaseMe.release();
                return makeRequest(contentType, content);
            }

            String reply = h.get("STATUSLINE").toString();
            
            if (reply.startsWith("407") || reply.startsWith("401")) {
                
                if (reply.startsWith("407")) doProxyAuth(h, content == null ? "GET" : "POST");
                else doWebAuth(h, content == null ? "GET" : "POST");
                
                if (h.get("HTTP").equals("1.0") && h.get("content-length") == null) {
                    if (Log.on) Log.log(this, "proxy returned an HTTP/1.0 reply with no content-length...");
                    in = null; sock = null;
                } else {
                    int cl = h.get("content-length") == null ? -1 : Integer.parseInt(h.get("content-length").toString());
                    new HTTPInputStream(in, cl, releaseMe).close();
                }
                releaseMe.release();
                return makeRequest(contentType, content);
                
            } else if (reply.startsWith("2")) {
                if (h.get("HTTP").equals("1.0") && h.get("content-length") == null)
                    throw new HTTPException("XWT does not support HTTP/1.0 servers which fail to return the Content-Length header");
                int cl = h.get("content-length") == null ? -1 : Integer.parseInt(h.get("content-length").toString());
                HTTPInputStream ret = new HTTPInputStream(in, cl, releaseMe);
                doRelease = false;
                return ret;
                
            } else {
                throw new HTTPException("HTTP Error: " + reply);
                
            }
            
        } catch (IOException e) { invalid = true; throw e;
        } finally { if (doRelease) releaseMe.release();
        }
    }


    // Safeguarded DNS Resolver ///////////////////////////////////////////////////////////////////////////

    /**
     *  resolves the hostname and returns it as a string in the form "x.y.z.w", except for the special case "xmlrpc.xwt.org".
     *  @throws HTTPException if the host falls within a firewalled netblock
     */
    private void resolveAndCheckIfFirewalled(String host) throws HTTPException {

        // special case
        if (host.equals("xmlrpc.xwt.org")) return;

        // cached
        if (resolvedHosts.get(host) != null) return;

        // if all scripts are trustworthy (local FS), continue
        if (Main.originAddr == null) return;

        // resolve using DNS
        try {
            InetAddress addr = InetAddress.getByName(host);
            byte[] quadbyte = addr.getAddress();
            if ((quadbyte[0] == 10 ||
                 (quadbyte[0] == 192 && quadbyte[1] == 168) ||
                 (quadbyte[0] == 172 && (quadbyte[1] & 0xF0) == 16)) && !addr.equals(Main.originAddr))
                throw new HTTPException("security violation: " + host + " [" + addr.getHostAddress() + "] is in a firewalled netblock");
            return;
        } catch (UnknownHostException uhe) { }

        // resolve using xmlrpc.xwt.org
        if (Platform.detectProxy() == null) throw new HTTPException("could not resolve hostname \"" + host + "\" and no proxy configured");
        if (Log.on) Log.log(this, "  could not resolve host " + host + "; using xmlrpc.xwt.org to ensure security");
        try {
            Object ret = new XMLRPC("http://xmlrpc.xwt.org/RPC2/", "dns.resolve").call(new Object[] { host });
            if (ret == null || !(ret instanceof String)) throw new Exception("    xmlrpc.xwt.org returned non-String: " + ret);
            resolvedHosts.put(host, ret);
            return;
        } catch (Throwable e) {
            throw new HTTPException("exception while attempting to use xmlrpc.xwt.org to resolve " + host + ": " + e);
        }
    }


    // Methods to attempt socket creation /////////////////////////////////////////////////////////////////

    /** Attempts a direct connection */
    public Socket attemptDirect() {
        try {
            if (Log.verbose) Log.log(this, "attempting to create unproxied socket to " + host + ":" + port + (ssl ? " [ssl]" : ""));
            return Platform.getSocket(host, port, ssl, true);
        } catch (IOException e) {
            if (Log.on) Log.log(this, "exception in attemptDirect(): " + e);
            return null;
        }
    }

    /** Attempts to use an HTTP proxy, employing the CONNECT method if HTTPS is requested */
    public Socket attemptHttpProxy(String proxyHost, int proxyPort) {
        try {
            if (Log.verbose) Log.log(this, "attempting to create HTTP proxied socket using proxy " + proxyHost + ":" + proxyPort);

            Socket sock = Platform.getSocket(proxyHost, proxyPort, ssl, false);
            if (!ssl) {
                if (!path.startsWith("http://")) path = "http://" + host + ":" + port + path;
            } else {
                PrintWriter pw = new PrintWriter(new OutputStreamWriter(sock.getOutputStream()));
                BufferedReader br = new BufferedReader(new InputStreamReader(sock.getInputStream()));
                pw.print("CONNECT " + host + ":" + port + " HTTP/1.1\r\n\r\n");
                pw.flush();
                String s = br.readLine();
                if (s.charAt(9) != '2') throw new HTTPException("proxy refused CONNECT method: \"" + s + "\"");
                while (br.readLine().length() > 0) { };
                ((TinySSL)sock).negotiate();
            }
            return sock;

        } catch (IOException e) {
            if (Log.on) Log.log(this, "exception in attemptHttpProxy(): " + e);
            return null;
        }
    }

    /**
     *  Implements SOCKSv4 with v4a DNS extension
     *  @see http://www.socks.nec.com/protocol/socks4.protocol
     *  @see http://www.socks.nec.com/protocol/socks4a.protocol
     */
    public Socket attemptSocksProxy(String proxyHost, int proxyPort) {

        // even if host is already a "x.y.z.w" string, we use this to parse it into bytes
        InetAddress addr = null;
        try { addr = InetAddress.getByName(host); } catch (Exception e) { }

        if (Log.verbose) Log.log(this, "attempting to create SOCKSv4" + (addr == null ? "" : "a") +
                                 " proxied socket using proxy " + proxyHost + ":" + proxyPort);

        try {
            Socket sock = Platform.getSocket(proxyHost, proxyPort, ssl, false);
            
            DataOutputStream dos = new DataOutputStream(sock.getOutputStream());
            dos.writeByte(0x04);                         // SOCKSv4(a)
            dos.writeByte(0x01);                         // CONNECT
            dos.writeShort(port & 0xffff);               // port
            if (addr == null) dos.writeInt(0x00000001);  // bogus IP
            else dos.write(addr.getAddress());           // actual IP
            dos.writeByte(0x00);                         // no userid
            if (addr == null) {
                PrintWriter pw = new PrintWriter(new OutputStreamWriter(dos));
                pw.print(host);
                pw.flush();
                dos.writeByte(0x00);                     // hostname null terminator
            }
            dos.flush();

            DataInputStream dis = new DataInputStream(sock.getInputStream());
            dis.readByte();                              // reply version
            byte success = dis.readByte();               // success/fail
            dis.skip(6);                                 // ip/port
            
            if ((int)(success & 0xff) == 90) {
                if (ssl) ((TinySSL)sock).negotiate();
                return sock;
            }
            if (Log.on) Log.log(this, "SOCKS server denied access, code " + (success & 0xff));
            return null;

        } catch (IOException e) {
            if (Log.on) Log.log(this, "exception in attemptSocksProxy(): " + e);
            return null;
        }
    }

    /** executes the PAC script and dispatches a call to one of the other attempt methods based on the result */
    public Socket attemptPAC(Function pacFunc) {
        if (Log.verbose) Log.log(this, "evaluating PAC script");
        String pac = null;
        try {
            Object obj = pacFunc.call(Context.enter(), Proxy.proxyAutoConfigRootScope, null, new Object[] { url.toString(), url.getHost() });
            if (Log.verbose) Log.log(this, "  PAC script returned \"" + obj + "\"");
            pac = obj.toString();
        } catch (Throwable e) {
            if (Log.on) Log.log(this, "PAC script threw exception " + e);
            return null;
        }

        StringTokenizer st = new StringTokenizer(pac, ";", false);
        while (st.hasMoreTokens()) {
            String token = st.nextToken().trim();
            if (Log.verbose) Log.log(this, "  trying \"" + token + "\"...");
            try {
                Socket ret = null;
                if (token.startsWith("DIRECT"))
                    ret = attemptDirect();
                else if (token.startsWith("PROXY"))
                    ret = attemptHttpProxy(token.substring(token.indexOf(' ') + 1, token.indexOf(':')),
                                           Integer.parseInt(token.substring(token.indexOf(':') + 1)));
                else if (token.startsWith("SOCKS"))
                    ret = attemptSocksProxy(token.substring(token.indexOf(' ') + 1, token.indexOf(':')),
                                            Integer.parseInt(token.substring(token.indexOf(':') + 1)));
                if (ret != null) return ret;
            } catch (Throwable e) {
                if (Log.on) Log.log(this, "attempt at \"" + token + "\" failed due to " + e + "; trying next token");
            }
        }
        if (Log.on) Log.log(this, "all PAC results exhausted");
        return null;
    }


    // Everything Else ////////////////////////////////////////////////////////////////////////////

    private synchronized void connect() throws IOException {
        if (sock != null) {
            if (in == null) in = new BufferedInputStream(sock.getInputStream());
            return;
        }
        // grab the userinfo; gcj doesn't have java.net.URL.getUserInfo()
        String url = originalUrl;
        userInfo = url.substring(url.indexOf("://") + 3);
        userInfo = userInfo.indexOf('/') == -1 ? userInfo : userInfo.substring(0, userInfo.indexOf('/'));
        if (userInfo.indexOf('@') != -1) {
            userInfo = userInfo.substring(0, userInfo.indexOf('@'));
            url = url.substring(0, url.indexOf("://") + 3) + url.substring(url.indexOf('@') + 1);
        } else {
            userInfo = null;
        }

        if (url.startsWith("https:")) {
            this.url = new URL("http" + url.substring(5));
            ssl = true;
        } else if (!url.startsWith("http:")) {
            throw new MalformedURLException("HTTP only supports http/https urls");
        } else {
            this.url = new URL(url);
        }
        if (!skipResolveCheck) resolveAndCheckIfFirewalled(this.url.getHost());
        port = this.url.getPort();
        path = this.url.getFile();
        if (port == -1) port = ssl ? 443 : 80;
        host = this.url.getHost();
        if (Log.verbose) Log.log(this, "creating HTTP object for connection to " + host + ":" + port);

        Proxy pi = Platform.detectProxy();
        if (sock == null && pi != null && pi.proxyAutoConfigFunction != null) sock = attemptPAC(pi.proxyAutoConfigFunction);
        if (sock == null && pi != null && ssl && pi.httpsProxyHost != null) sock = attemptHttpProxy(pi.httpsProxyHost, pi.httpsProxyPort);
        if (sock == null && pi != null && pi.httpProxyHost != null) sock = attemptHttpProxy(pi.httpProxyHost, pi.httpProxyPort);
        if (sock == null && pi != null && pi.socksProxyHost != null) sock = attemptSocksProxy(pi.socksProxyHost, pi.socksProxyPort);
        if (sock == null) sock = attemptDirect();
        if (sock == null) throw new HTTPException("unable to contact host " + host);
        if (in == null) in = new BufferedInputStream(sock.getInputStream());
    }

    public void sendRequest(String contentType, String content) throws IOException {

        PrintWriter pw = new PrintWriter(new OutputStreamWriter(sock.getOutputStream()));
        if (content != null) {
            pw.print("POST " + path + " HTTP/1.1\r\n");
	    int contentLength = content.substring(0, 2).equals("\r\n") ?
		content.length() - 2 :
		(content.length() - content.indexOf("\r\n\r\n") - 4);
            pw.print("Content-Length: " + contentLength + "\r\n");
            if (contentType != null) pw.print("Content-Type: " + contentType + "\r\n");
        } else {
            pw.print("GET " + path + " HTTP/1.1\r\n");
        }
        
        pw.print("User-Agent: XWT\r\n");
        pw.print("Host: " + (host + (port == 80 ? "" : (":" + port))) + "\r\n");

        if (Proxy.Authorization.authorization != null) pw.print("Proxy-Authorization: " + Proxy.Authorization.authorization2 + "\r\n");
        if (authCache.get(originalUrl) != null) pw.print("Authorization: " + authCache.get(originalUrl) + "\r\n");

        pw.print(content == null ? "\r\n" : content);
        pw.print("\r\n");
        pw.flush();
    }

    private void doWebAuth(Hashtable h0, String method) throws IOException {
        if (userInfo == null) throw new HTTPException("web server demanded username/password, but none were supplied");
        Hashtable h = parseAuthenticationChallenge(h0.get("www-authenticate").toString());
        
        if (h.get("AUTHTYPE").equals("Basic")) {
            if (authCache.get(originalUrl) != null) throw new HTTPException("username/password rejected");
            authCache.put(originalUrl, "Basic " + new String(Base64.encode(userInfo.getBytes("US-ASCII"))));
            
        } else if (h.get("AUTHTYPE").equals("Digest")) {
            if (authCache.get(originalUrl) != null && !"true".equals(h.get("stale"))) throw new HTTPException("username/password rejected");
            String path2 = path;
            if (path2.startsWith("http://") || path2.startsWith("https://")) {
                path2 = path2.substring(path2.indexOf("://") + 3);
                path2 = path2.substring(path2.indexOf('/'));
            }
            String A1 = userInfo.substring(0, userInfo.indexOf(':')) + ":" + h.get("realm") + ":" + userInfo.substring(userInfo.indexOf(':') + 1);
            String A2 = method + ":" + path2;
            authCache.put(originalUrl,
                          "Digest " +
                          "username=\"" + userInfo.substring(0, userInfo.indexOf(':')) + "\", " +
                          "realm=\"" + h.get("realm") + "\", " +
                          "nonce=\"" + h.get("nonce") + "\", " +
                          "uri=\"" + path2 + "\", " +
                          (h.get("opaque") == null ? "" : ("opaque=\"" + h.get("opaque") + "\", ")) + 
                          "response=\"" + H(H(A1) + ":" + h.get("nonce") + ":" + H(A2)) + "\", " +
                          "algorithm=MD5"
                          );
            
        } else {
            throw new HTTPException("unknown authentication type: " + h.get("AUTHTYPE"));
        }
    }

    private void doProxyAuth(Hashtable h0, String method) throws IOException {
        if (Log.on) Log.log(this, "Proxy AuthChallenge: " + h0.get("proxy-authenticate"));
        Hashtable h = parseAuthenticationChallenge(h0.get("proxy-authenticate").toString());
        String style = h.get("AUTHTYPE").toString();
        String realm = h.get("realm").toString();

        if (!realm.equals("Digest") || Proxy.Authorization.authorization2 == null || !"true".equals(h.get("stale")))
            Proxy.Authorization.getPassword(realm, style, sock.getInetAddress().getHostAddress(), Proxy.Authorization.authorization);

        if (style.equals("Basic")) {
            Proxy.Authorization.authorization2 =
                "Basic " + new String(Base64.encode(Proxy.Authorization.authorization.getBytes("US-ASCII")));
            
        } else if (style.equals("Digest")) {
            String A1 = Proxy.Authorization.authorization.substring(0, userInfo.indexOf(':')) + ":" + h.get("realm") + ":" +
                Proxy.Authorization.authorization.substring(Proxy.Authorization.authorization.indexOf(':') + 1);
            String A2 = method + ":" + path;
            Proxy.Authorization.authorization2 = 
                "Digest " +
                "username=\"" + Proxy.Authorization.authorization.substring(0, Proxy.Authorization.authorization.indexOf(':')) + "\", " +
                "realm=\"" + h.get("realm") + "\", " +
                "nonce=\"" + h.get("nonce") + "\", " +
                "uri=\"" + path + "\", " +
                (h.get("opaque") == null ? "" : ("opaque=\"" + h.get("opaque") + "\", ")) + 
                "response=\"" + H(H(A1) + ":" + h.get("nonce") + ":" + H(A2)) + "\", " +
                "algorithm=MD5";
        }            
    }


    // HTTPException ///////////////////////////////////////////////////////////////////////////////////

    static class HTTPException extends IOException {
        public HTTPException(String s) { super(s); }
    }


    // HTTPInputStream ///////////////////////////////////////////////////////////////////////////////////

    /** An input stream that represents a subset of a longer input stream. Supports HTTP chunking as well */
    public class HTTPInputStream extends FilterInputStream {

        /** if chunking, the number of bytes remaining in this subset; otherwise the remainder of the chunk */
        private int length = 0;

        /** this semaphore will be released when the stream is closed */
        private Semaphore releaseMe = null;

        /** indicates that we have encountered the zero-length terminator chunk */
        boolean chunkedDone = false;

        /** if we're on the first chunk, we don't pre-read a CRLF */
        boolean firstChunk = true;

        /** the length of the entire content body; -1 if chunked */
        private int contentLength = 0;
        public int getContentLength() { return contentLength; }

        HTTPInputStream(InputStream in, int length, Semaphore releaseMe) {
            super(in);
            this.releaseMe = releaseMe;
            this.contentLength = length;
            this.length = length == -1 ? 0 : length;
        }

        private void readChunk() throws IOException {
            if (chunkedDone) return;
            if (!firstChunk) super.skip(2); // CRLF
            firstChunk = false;
            String chunkLen = "";
            while(true) {
                int i = super.read();
                if (i == -1) throw new HTTPException("encountered end of stream while reading chunk length");

                // FIXME: handle chunking extensions
                if (i == '\r') {
                    super.read();    // LF
                    break;
                } else {
                    chunkLen += (char)i;
                }
            }
            length = Integer.parseInt(chunkLen, 16);
            if (length == 0) chunkedDone = true;
        }

        public int read(byte[] b, int off, int len) throws IOException {
            boolean good = false;
            try {
                if (length == 0 && contentLength == -1) readChunk();
                if (len > length) len = length;
                int ret = super.read(b, off, len);
                length -= ret;
                good = true;
                return ret;
            } finally {
                if (!good) invalid = true;
            }
        }

        public void close() throws IOException {
            if (contentLength == -1) {
                while(!chunkedDone) {
                    if (length != 0) skip(length);
                    readChunk();
                }
                skip(2);
            } else {
                skip(length);
            }
            if (releaseMe != null) releaseMe.release();
        }
    }


    // Misc Helpers ///////////////////////////////////////////////////////////////////////////////////

    /** reads a set of HTTP headers off of the input stream, returning null if the stream is already at its end */
    private Hashtable parseHeaders(InputStream in) throws IOException {
        Hashtable ret = new Hashtable();

        // we can't use a BufferedReader directly on the input stream, since it will buffer past the end of the headers
        byte[] buf = new byte[4096];
        int buflen = 0;
        while(true) {
            int read = in.read();
            if (read == -1 && buflen == 0) return null;
            if (read == -1) throw new HTTPException("stream closed while reading headers");
            buf[buflen++] = (byte)read;
            if (buflen >= 4 && buf[buflen - 4] == '\r' && buf[buflen - 3] == '\n' && buf[buflen - 2] == '\r' && buf[buflen - 1] == '\n') break;
            if (buflen == buf.length) {
                byte[] newbuf = new byte[buf.length * 2];
                System.arraycopy(buf, 0, newbuf, 0, buflen);
                buf = newbuf;
            }
        }

        BufferedReader br = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(buf, 0, buflen)));
        String s = br.readLine();
        if (!s.startsWith("HTTP/")) throw new HTTPException("Expected reply to start with \"HTTP/\", got: " + s);
        ret.put("STATUSLINE", s.substring(s.indexOf(' ') + 1));
        ret.put("HTTP", s.substring(5, s.indexOf(' ')));

        while((s = br.readLine()) != null && s.length() > 0) {
            String front = s.substring(0, s.indexOf(':')).toLowerCase();
            String back = s.substring(s.indexOf(':') + 1).trim();
            // ugly hack: we never replace a Digest-auth with a Basic-auth (proxy + www)
            if (front.endsWith("-authenticate") && ret.get(front) != null && !back.equals("Digest")) continue;
            ret.put(front, back);
        }
        return ret;
    }

    private Hashtable parseAuthenticationChallenge(String s) {
        Hashtable ret = new Hashtable();

        s = s.trim();
        ret.put("AUTHTYPE", s.substring(0, s.indexOf(' ')));
        s = s.substring(s.indexOf(' ')).trim();

        while (s.length() > 0) {
            String val = null;
            String key = s.substring(0, s.indexOf('='));
            s = s.substring(s.indexOf('=') + 1);
            if (s.charAt(0) == '\"') {
                s = s.substring(1);
                val = s.substring(0, s.indexOf('\"'));
                s = s.substring(s.indexOf('\"') + 1);
            } else {
                val = s.indexOf(',') == -1 ? s : s.substring(0, s.indexOf(','));
                s = s.indexOf(',') == -1 ? "" : s.substring(s.indexOf(',') + 1);
            }
            if (s.length() > 0 && s.charAt(0) == ',') s = s.substring(1);
            s = s.trim();
            ret.put(key, val);
        }
        return ret;
    }

    private String H(String s) throws IOException {
        byte[] b = s.getBytes("US-ASCII");
        MD5Digest md5 = new MD5Digest();
        md5.update(b, 0, b.length);
        byte[] out = new byte[md5.getDigestSize()];
        md5.doFinal(out, 0);
        String ret = "";
        for(int i=0; i<out.length; i++) {
            ret += "0123456789abcdef".charAt((out[i] & 0xf0) >> 4);
            ret += "0123456789abcdef".charAt(out[i] & 0x0f);
        }
        return ret;
    }

}
